import numpy as np

import rpy2.robjects as ro
from rpy2.robjects.packages import importr

from tbats.tbats import Model as TBATSModel


class RComparisonBase(object):

    def r_bats(self, y, components):
        components = components.copy()
        if 'seasonal_periods' in components:
            components['seasonal_periods'] = ro.IntVector(components['seasonal_periods'])
        importr('forecast')
        r_bats_func = ro.r['bats']
        r_forecast = ro.r['forecast']
        r_y = ro.FloatVector(list(y))
        r_model = r_bats_func(r_y, **components)
        summary = r_forecast(r_model)
        # predictions = np.array(summary.rx('fitted')).flatten()
        return summary, r_model

    def r_tbats(self, y, components):
        components = components.copy()
        if 'seasonal_periods' in components:
            components['seasonal_periods'] = ro.IntVector(components['seasonal_periods'])
        importr('forecast')
        r_bats_func = ro.r['tbats']
        r_forecast = ro.r['forecast']
        r_y = ro.FloatVector(list(y))
        r_model = r_bats_func(r_y, **components)
        summary = r_forecast(r_model)
        # predictions = np.array(summary.rx('fitted')).flatten()
        return summary, r_model

    def assert_py_model_is_not_worse(self, y, r_summary, r_model, py_model):
        r_predictions = np.asarray(r_summary.rx('fitted')).flatten()
        r_aic = self.get_R_float_or_null(r_model, 'AIC')

        aic = py_model.aic

        if r_aic < aic:
            assert np.isclose(r_aic, aic, atol=0.1)

        r_resid = np.asarray(r_predictions) - np.asarray(y)
        r_mae = np.abs(r_resid).mean()
        mae = np.abs(py_model.resid).mean()

        if r_mae < mae:
            assert np.isclose(r_mae, mae, atol=0.1)

    def assert_forecast_is_not_worse(self, y_to_forecast, r_model, py_model):

        steps = len(y_to_forecast)

        # Forecast from R model
        r_forecast = ro.r['forecast']
        r_predictions = r_forecast(r_model, h=steps)
        r_y_hat = np.array(r_predictions.rx('mean')).flatten()

        # Forecast from python model
        py_y_hat = py_model.forecast(steps)

        r_resid = r_y_hat - y_to_forecast
        py_resid = py_y_hat - y_to_forecast

        r_mae = np.abs(r_resid).mean()
        py_mae = np.abs(py_resid).mean()

        if r_mae < py_mae:
            assert np.isclose(r_mae, py_mae, atol=0.01)

    def compare_model(self, r_summary, r_model, py_model, atol_small=0.001, atol_big=0.01, atol_for_series=0.001):
        r_predictions = np.array(r_summary.rx('fitted')).flatten()

        r_alpha = self.get_R_float_or_null(r_model, 'alpha')
        r_beta = self.get_R_float_or_null(r_model, 'beta')
        r_damping_parameter = self.get_R_float_or_null(r_model, 'damping.parameter')
        r_box_cox_lambda = self.get_R_float_or_null(r_model, 'lambda')
        r_aic = self.get_R_float_or_null(r_model, 'AIC')
        r_gamma = self.get_R_array_or_null(r_model, 'gamma.values')
        r_gamma_1 = self.get_R_array_or_null(r_model, 'gamma.one.values')
        r_gamma_2 = self.get_R_array_or_null(r_model, 'gamma.two.values')
        r_seasonal_periods = self.get_R_array_or_null(r_model, 'seasonal.periods')
        r_ar_coefs = self.get_R_array_or_null(r_model, 'ar.coefficients')
        r_ma_coefs = self.get_R_array_or_null(r_model, 'ma.coefficients')
        r_k_vector = self.get_R_array_or_null(r_model, 'k.vector')

        r_x0 = np.array((r_model.rx('seed.states')[0])).flatten()

        aic = py_model.aic

        # simple parameters comparison
        self.assert_none_or_close(r_alpha, py_model.params.alpha, atol=atol_small)
        self.assert_none_or_close(r_beta, py_model.params.beta, atol=atol_small)
        self.assert_none_or_close(r_damping_parameter, py_model.params.phi, atol=atol_small)
        self.assert_none_or_close(r_box_cox_lambda, py_model.params.box_cox_lambda, atol=atol_small)

        self.assert_array_or_close(r_seasonal_periods, py_model.params.components.seasonal_periods)
        if isinstance(py_model, TBATSModel):
            self.assert_array_or_close(r_k_vector, py_model.params.components.seasonal_harmonics)
            self.assert_array_or_close(r_gamma_1, py_model.params.gamma_1(), atol=atol_big)
            self.assert_array_or_close(r_gamma_2, py_model.params.gamma_2(), atol=atol_big)
        else:
            self.assert_array_or_close(r_gamma, py_model.params.gamma_params)

        self.assert_array_or_close(r_ar_coefs, py_model.params.ar_coefs, atol=atol_big)
        self.assert_array_or_close(r_ma_coefs, py_model.params.ma_coefs, atol=atol_big)

        # seed states comparison
        assert np.allclose(r_x0, py_model.params.x0, atol=atol_small)

        # predictions comparison
        assert np.allclose(r_predictions, py_model.y_hat, atol=atol_for_series)

        # metrics comparison
        assert np.isclose(r_aic, aic, atol=atol_small)

    def compare_forecast(self, r_model, py_model, steps=20, atol=0.01):

        # Forecast from R model
        r_forecast = ro.r['forecast']
        r_predictions = r_forecast(r_model, h=steps)
        r_y_hat = np.array(r_predictions.rx('mean')).flatten()

        # Forecast from python model
        py_y_hat = py_model.forecast(steps)

        assert np.allclose(r_y_hat, py_y_hat, atol=atol)

    def assert_array_or_close(self, r_value, py_value, atol=0.001):
        if r_value is None:
            assert len(py_value) == 0
        else:
            assert np.allclose(r_value, py_value, atol=atol)

    def assert_none_or_close(self, r_value, py_value, atol=0.001):
        if r_value is None:
            assert py_value is None
        else:
            assert np.isclose(r_value, py_value, atol=atol)

    def get_R_float_or_null(self, r_model, paramName):
        val = np.array(r_model.rx(paramName)).flatten()[0]
        if isinstance(val, ro.rinterface.RNULLType):
            return None
        return float(val)

    def get_R_array_or_null(self, r_model, paramName):
        val = np.array(r_model.rx(paramName)).flatten()
        if len(val) == 0 or isinstance(val[0], ro.rinterface.RNULLType):
            return None
        return val



import numpy as np

from tbats import TBATS
from test_R.RComparisonBase import RComparisonBase


class TestTBATSComparisonR(RComparisonBase):
    """
    In order to run those tests you need installed R and it's forecast package
    """

    def test_damped_trend(self):
        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=True,
            use_box_cox=False
        )

        alpha = 0.4
        beta = 0.6
        phi = 0.9
        np.random.seed(987)
        T = 100

        b = 0
        b_long = 0.0
        l = 1
        y = [0] * T
        for t in range(0, T):
            d = np.random.normal(scale=1.0)
            y[t] = l + b + d
            l = l + b + alpha * d
            b = (1 - phi) * b_long + phi * b + beta * d

        r_summary, r_model = self.r_tbats(y, components)

        estimator = TBATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model)
        self.compare_forecast(r_model, py_model)

    def test_trend_and_seasonal(self):
        np.random.seed(234234)
        T = 35
        steps = 5
        alpha = 0.1
        period_length = 6
        y = [0] * T
        b = b0 = 2.1
        l = l0 = 1.2
        for t in range(0, T):
            d = np.random.normal()
            y[t] = l + b + d + 2 * np.sin(2 * np.pi * t / period_length)
            l = l + b + alpha * d

        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=False,
            use_box_cox=False,
            seasonal_periods=[period_length],
        )

        y_for_train = y[:(T - steps)]
        y_to_forecast = y[(T - steps):]

        r_summary, r_model = self.r_tbats(y_for_train, components)

        estimator = TBATS(**components)
        py_model = estimator.fit(y_for_train)

        self.assert_py_model_is_not_worse(y_for_train, r_summary, r_model, py_model)
        self.assert_forecast_is_not_worse(y_to_forecast, r_model, py_model)

    def test_trend_and_two_seasonalities(self):
        np.random.seed(234234)
        T = 35
        steps = 5
        alpha = 0.1
        period_1_length = 3
        period_2_length = 7
        y = [0] * T
        b = b0 = 2.1
        l = l0 = 1.2
        for t in range(0, T):
            d = np.random.normal()
            s1 = 2 * np.cos(2 * np.pi * t / period_1_length)
            s2 = 3 * np.sin(2 * np.pi * t / period_2_length)
            y[t] = l + b + s1 + s2 + d
            l = l + b + alpha * d

        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=False,
            use_box_cox=False,
            seasonal_periods=[period_1_length, period_2_length],
        )

        y_for_train = y[:(T - steps)]
        y_to_forecast = y[(T - steps):]

        r_summary, r_model = self.r_tbats(y_for_train, components)

        estimator = TBATS(**components)
        py_model = estimator.fit(y_for_train)

        self.assert_py_model_is_not_worse(y_for_train, r_summary, r_model, py_model)
        self.assert_forecast_is_not_worse(y_to_forecast, r_model, py_model)

    def test_long_seasonality(self):
        np.random.seed(5434)
        T = 300
        steps = 5
        alpha = 0.1
        period_1_length = 7
        period_2_length = 30.5
        y = [0] * T
        b = b0 = 2.1
        l = l0 = 1.2
        for t in range(0, T):
            d = np.random.normal()
            s1 = 2 * np.cos(2 * np.pi * t / period_1_length)
            s2 = 3 * np.sin(2 * np.pi * t / period_2_length)
            y[t] = l + b + s1 + s2 + d
            l = l + b + alpha * d

        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=False,
            use_box_cox=False,
            seasonal_periods=[period_1_length, period_2_length],
        )

        y_for_train = y[:(T - steps)]

        y_to_forecast = y[(T - steps):]

        r_summary, r_model = self.r_tbats(y_for_train, components)

        estimator = TBATS(n_jobs=1, **components)
        py_model = estimator.fit(y_for_train)

        self.assert_py_model_is_not_worse(y_for_train, r_summary, r_model, py_model)
        self.assert_forecast_is_not_worse(y_to_forecast, r_model, py_model)


import numpy as np

from tbats import BATS
from .RComparisonBase import RComparisonBase


class TestBATSComparisonR(RComparisonBase):
    """
    In order to run those tests you need installed R and it's forecast package
    """

    def test_damped_trend(self):
        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=True,
            use_box_cox=False
        )

        alpha = 0.4
        beta = 0.6
        phi = 0.9
        np.random.seed(987)
        T = 100

        b = 0
        b_long = 0.0
        l = 1
        y = [0] * T
        for t in range(0, T):
            d = np.random.normal(scale=1.0)
            y[t] = l + b + d
            l = l + b + alpha * d
            b = (1 - phi) * b_long + phi * b + beta * d

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model)
        self.compare_forecast(r_model, py_model)

    def test_simple_seasonality(self):
        season_length = 5
        components = dict(
            use_arma_errors=False,
            use_trend=False,
            use_box_cox=False,
            seasonal_periods=[season_length],
        )

        np.random.seed(3244)
        gamma = 0.4
        seasonal_starting_params = [-0.5, 0.5, 0.8, -0.8, 0.0]  # sum is 0
        T = 50

        l = 1
        y = [0] * T
        s = [0] * (T + season_length)
        s[0:season_length] = seasonal_starting_params
        for t in range(0, T):
            d = np.random.normal()
            s[t + season_length] = s[t] + gamma * d
            y[t] = l + s[t + season_length] + d

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model)
        self.compare_forecast(r_model, py_model)

    def test_seasonality_with_subseasonality(self):
        season_length = 4

        components = dict(
            use_arma_errors=False,
            use_trend=False,
            use_box_cox=False,
            seasonal_periods=[2, season_length],
        )

        np.random.seed(3244)

        gamma = 0.6
        seasonal_starting_params = [-0.2, 0.2, -0.8, 0.8]  # sum is 0
        T = 50

        l = 1
        y = [0] * T
        s = [0] * (T + season_length)
        s[0:season_length] = seasonal_starting_params
        for t in range(0, T):
            d = np.random.normal()
            s[t + season_length] = s[t] + gamma * d
            y[t] = l + s[t + season_length] + d

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model)
        self.compare_forecast(r_model, py_model)

    def test_2_seasonalities_with_common_divisor(self):
        season_length_1 = 4
        season_length_2 = 6  # common divisor is 2

        components = dict(
            use_arma_errors=False,
            use_trend=False,
            use_box_cox=False,
            seasonal_periods=[season_length_1, season_length_2],
        )

        np.random.seed(3214)
        gamma1 = 0.6
        gamma2 = 0.5
        seasonal_starting_params_1 = [-0.2, 0.2, -0.8, 0.8]  # sum is 0
        seasonal_starting_params_2 = [-0.3, 0.2, 0.6, -0.6, 0.3, -0.2]  # sum is 0
        T = 300

        l = 1
        y = [0] * T
        s1 = [0] * (T + season_length_1)
        s1[0:season_length_1] = seasonal_starting_params_1
        s2 = [0] * (T + season_length_2)
        s2[0:season_length_2] = seasonal_starting_params_2
        for t in range(0, T):
            d = np.random.normal()
            s1[t + season_length_1] = s1[t] + gamma1 * d
            s2[t + season_length_2] = s2[t] + gamma2 * d
            y[t] = l + s1[t + season_length_1] + s1[t + season_length_1] + d

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model, atol_small=0.005, atol_for_series=0.005)
        self.compare_forecast(r_model, py_model, atol=0.005)

    def test_arma_errors(self):
        # Note: this test is very fragile to optimization function changes
        # as it is not easy to converge those series with ARMA errors.
        components = dict(
            use_arma_errors=True,
            use_trend=False,
            use_box_cox=False,
        )

        np.random.seed(3244)
        alpha = 0.9
        T = 100

        l = 1
        e = 0
        e_prev = 0
        e_prev_2 = 0
        y = [0] * T
        for t in range(0, T):
            e_prev_3 = e_prev_2
            e_prev_2 = e_prev
            e_prev = e
            e = np.random.normal()
            d = e + 0.1 * e_prev - 0.5 * e_prev_2 + 0.3 * e_prev_3  # AR(3) for errors
            y[t] = l + alpha * d

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        assert py_model.params.components.arma_length() > 0  # there are ARMA components in the final model
        self.compare_model(r_summary, r_model, py_model, atol_small=0.01, atol_for_series=0.01)
        self.compare_forecast(r_model, py_model)

    def test_fit_boxcox(self):
        components = dict(
            use_arma_errors=False,
            use_trend=True,
            use_damped_trend=False,
            use_box_cox=True,
        )

        np.random.seed(49385)
        y = np.exp(
            np.random.uniform(size=100) + np.array(range(40, 140)) / 20
        )  # variance is increasing in those series

        r_summary, r_model = self.r_bats(y, components)

        estimator = BATS(**components)
        py_model = estimator.fit(y)

        self.compare_model(r_summary, r_model, py_model, atol_for_series=0.2)
        self.compare_forecast(r_model, py_model, atol=0.2)
